/*----------------------------------------------------------------------------*\
|                                                                              |
| Copyright (C) 2001-2002,  James A. Cureington                                |
|                           tonycu at users_sourceforge_net                    |
|                                                                              |
| This program is free software; you can redistribute it and/or                |
| modify it under the terms of the GNU General Public License                  |
| as published by the Free Software Foundation; either version 2               |
| of the License, or any later version.                                        |
|                                                                              |
| This program is distributed in the hope that it will be useful,              |
| but WITHOUT ANY WARRANTY; without even the implied warranty of               |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                |
| GNU General Public License for more details.                                 |
|                                                                              |
| You should have received a copy of the GNU General Public License            |
| along with this program; if not, write to the Free Software                  |
| Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                    |
|                                                                              |
|                                                                              |
| This driver downloads USB boards built with the Cypress (formely Anchor      |
| Chips EZ-USB AN2131 chip. While it should download any board based on the    |
| above chip, I've only tested it on the USBSIMM board by J. Gordon            |
| Electronics (www.jged.com). I've tested this with two USBSIMM boards         |
| connected at the same time.                                                  |
|                                                                              |
| There is no hard coded limit for the number of boards that may be connected  |
| in this driver - all structure allocation is dynamic and should be SMP safe. |
|                                                                              |
| See the README file for up-to-date info on building this driver, downloading |
| connected USBSIMM boards, and a list of USB resources.                       |
|                                                                              |
|                                                                              |
\*----------------------------------------------------------------------------*/

/******************************************************************************\
* $Id: ezusb2131.c,v 1.7 2003/02/21 06:34:42 tonycu Exp $
*
* $Log: ezusb2131.c,v $
* Revision 1.7  2003/02/21 06:34:42  tonycu
* updated version to 1.0
*
* Revision 1.6  2003/02/08 15:49:23  tonycu
* Change lock flags from int to unsigned long per Martin Frey
* martin_frey * ei_tum_de
*
* Revision 1.5  2002/06/05 06:44:06  tonycu
*
* A special hack was added for developers when the device
* re-enumerates as an EZUSB device (venid=0x547, prodid=0x2131)
* after it has been downloaded by the EZUSB2131 driver. In this
* case the developer does not want the EZUSB2131 driver to claim
* the device after it re-enumerates. This is achieved by the
* ignore_renum_modula module parameter, ignore_renum_modula forces
* the re-enumeration to be ignored when
* (number_of_connects%ignore_renum_modula)==0. The number_of_connects
* variable increases with each ezusb device connection. So, passing
* ignore_renum_modula=2 as a module parameter would cause every other
* ezusb connection to be ignored. If the ezusb device is ignored, a
* message stating so will be logged in /var/log/messages.
*
* Revision 1.4  2002/01/20 18:02:23  tonycu
* update version...
*
* Revision 1.3  2002/01/20 17:57:49  tonycu
* removed #include <linux/smp_lock.h>, it caused compile failure on
* 2.4.7-10 SMP kernels...
*
* Revision 1.2  2001/07/22 20:09:26  tonycu
* Incorporate contributions from Arnim Laeuger (arnim.laeuger at gmx_net)
* to allow user to specify non-native vendor and product id as module
* parameters.
*
* Revision 1.1  2001/05/30 05:29:53  tonycu
* initial check-in...
*
*
\******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif /* MODVERSIONS */
#include <linux/module.h>

#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
/* include below causes problems on 2.4.7-10 SMP... 
#include <linux/smp_lock.h>
*/
#include <linux/spinlock.h>
#include <linux/socket.h>
#include <linux/miscdevice.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>


/* module stuff ***************************************************************/
        /* must use define below to use in MODULE_* below; static const char */
        /* strings won't work... just if you'r wondering why i used #define  */
#define EZUSB_DESCRIPTION "EZ-USB AN2131 Board Loader"
#define EZUSB_VERSION     "1.0"
#define EZUSB_COPYRIGHT   "ezusb2131 driver Copyright (c)2001-2002, Tony Cureington <tonycu at users_sourceforge_net>"

#define ANCHOR_CHIPS_VENID    0x0547 /* Anchor Chips Inc.                   */
#define AN2131_DEVID          0x2131 /* AN2131 EZUSB Microcontroller        */

static int uid=0;            /* root is the default owner of */
                             /* /proc/... download files     */

static int multi_boards=0;   /* if downloading a multiple boards we   */
                             /* must base the download file name on   */
                             /* the dynamic usb device number; if set */
                             /* to 1, the download file is always     */
                             /* /proc/ezusb/dev0                      */

static int vendor_id=ANCHOR_CHIPS_VENID;       /* default vendor id   */
static int product_id=AN2131_DEVID;            /* default product id  */

static int ignore_renum_modula=INT_MAX;  /* force device connects to be */
                                         /* ignored when num_connect%   */
                                         /* ignore_renum_modula == 0    */

MODULE_AUTHOR("Tony Cureington; tonycu at users_sourceforge_net");
MODULE_DESCRIPTION(EZUSB_DESCRIPTION ", Version " 
                   EZUSB_VERSION"; " EZUSB_COPYRIGHT);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#ifdef MODULE_SUPPORTED_DEVICE
MODULE_SUPPORTED_DEVICE("dev/proto boards based on EZ-USB AN2131 chips");
#endif
MODULE_PARM(uid, "i");
MODULE_PARM_DESC(uid, "user id of user allowed to download boards");
MODULE_PARM(multi_boards, "i");
MODULE_PARM_DESC(multi_boards, "if downloading multiple boards, set to 1");
MODULE_PARM(vendor_id, "i");
MODULE_PARM_DESC(vendor_id, "vendor id for a non-native device");
MODULE_PARM(product_id, "i");
MODULE_PARM_DESC(product_id, "product id for a non-native device");
MODULE_PARM(ignore_renum_modula, "i");
MODULE_PARM_DESC(ignore_renum_modula, "ignore device re-enumeration modula, see EZUSB2131 driver README for details");


/* defines ********************************************************************/
#define EZUSB_DEVICE_TABLE    1    /* use MODULE_DEVICE_TABLE, set to zero if */
                                   /* you don't want the usb subsystem to     */
                                   /* auto load this driver... but usb subsys */
                                   /* will complain about not being able to   */
                                   /* find the driver even if already loaded  */
                                   /* by insmod...                            */

#define EZUSB_MOD_NAME        "ezusb2131"  /* module name */

#define EZUSB_BUFF_LEN        120
#define EZUSB_PROCFS_PERMS    (S_IFREG | S_IWUSR)

#define CPUCS_REG             0x7F92 /* the CPUCS is the only register that */
                                     /* can be written to using the Anchor  */
                                     /* download command...                 */

#define CPUCS_RESET              0x1
#define CPUCS_RUN                0x0

#define VEN_REQUEST_OUT          0x40   /* vendor request out */
#define VEN_REQUEST_IN           0xC0   /* vendor request in  */
#define ANCHOR_LOAD              0xA0   /* vendor request out */

#define EZUSB_PROC_FILE_OPENED   0x01   /* state for use with */
#define EZUSB_DOWNLOADING        0x02   /* procfs             */
#define EZUSB_DOWNLOAD_TIMEOUT   0x04
#define EZUSB_REMOVE_PENDING     0x08

#define EZUSB_LOAD_TIMEOUT       (HZ*3) /* three seconds, good enough eh? */ 

#define EZUSB_LL KERN_ALERT /* ezusb log level, change to level you desire */

#define MAX_IHEX_RECORD_LEN 16  /* intel hex record length */


/* marcros ********************************************************************/

        /* these two LOCK macros take into account if we are in an */
        /* interrupt thread; we use spin_lock if in an interrupt   */
        /* thread and spin_lock_irqsave otherwise; this prevents   */
        /* getting into a dead-lock if we are holding a lock when  */
        /* an interrupt occurs that contends for the same lock.    */
        /* ok, ok, so a cheat a bit on these lock macros... this   */
        /* does not have to be fast code, so I take the easy route */
#define EZUSB_SPIN_LOCK(lock, flags)            \
            if (in_interrupt())                 \
            {                                   \
               spin_lock(lock);                 \
            }                                   \
            else                                \
            {                                   \
               spin_lock_irqsave(lock, flags);  \
            }

#define EZUSB_SPIN_UNLOCK(lock, flags)               \
            if (in_interrupt())                      \
            {                                        \
               spin_unlock(lock);                    \
            }                                        \
            else                                     \
            {                                        \
               spin_unlock_irqrestore(lock, flags);  \
            }

        /* these macros are similar to the above macro but provide for read */
        /* or read/write locking to allow multiple readers in the same code */
#define EZUSB_READ_LOCK(lock, flags)            \
            if (in_interrupt())                 \
            {                                   \
               read_lock(lock);                 \
            }                                   \
            else                                \
            {                                   \
               read_lock_irqsave(lock, flags);  \
            }
 
#define EZUSB_READ_UNLOCK(lock, flags)               \
            if (in_interrupt())                      \
            {                                        \
               read_unlock(lock);                    \
            }                                        \
            else                                     \
            {                                        \
               read_unlock_irqrestore(lock, flags);  \
            }

#define EZUSB_WRITE_LOCK(lock, flags)            \
            if (in_interrupt())                  \
            {                                    \
               write_lock(lock);                 \
            }                                    \
            else                                 \
            {                                    \
               write_lock_irqsave(lock, flags);  \
            }
 
#define EZUSB_WRITE_UNLOCK(lock, flags)               \
            if (in_interrupt())                       \
            {                                         \
               write_unlock(lock);                    \
            }                                         \
            else                                      \
            {                                         \
               write_unlock_irqrestore(lock, flags);  \
            }


/* externs ********************************************************************/
extern unsigned long volatile jiffies;
extern unsigned int local_irq_count[NR_CPUS];
extern unsigned int local_bh_count[NR_CPUS];


/* structures *****************************************************************/
typedef struct
{
   unsigned char   length;
   unsigned short  address;
   unsigned char   type;
   unsigned char   data[MAX_IHEX_RECORD_LEN];
} intel_hex_record_t;


typedef struct _ezusb_elem_t
{
   unsigned int device_num;      /* the device number, or set to zero if   */
                                 /* multi_boards is st to one; this allows */
                                 /* the user to download to the same file  */
   char   buffer[EZUSB_BUFF_LEN];
   int    buffer_bytes_used;
   struct timer_list download_timer;

   int bytes_downloaded;
   int state;
   spinlock_t state_lock;  /* used to prevent race conditions when setting   */
                           /* and testing bits in state above... no need to  */
                           /* use read/write lock because most of the time   */
                           /* we test and set... so just use spin_lock...    */

   struct semaphore   mutex;
   struct usb_device *usbdev;
   wait_queue_head_t  remove_ok;
   struct file *procfs_file;    /* used to match proc file with usb dev */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
   struct proc_dir_entry *procfs_entry;
#else
   struct proc_dir_entry procfs_entry;
#endif

   struct _ezusb_elem_t *next;
      
} ezusb_elem_t;


typedef struct
{
   rwlock_t list_lock;     /* used to lock list below, we use read/write     */
                           /* lock to allow of multiple readers to search    */
                           /* the list for elements concurrently             */
   ezusb_elem_t *ezusb_list;
} ezusb_t;


/* local function declarations ************************************************/
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
static void *ezusb_probe(struct usb_device *, unsigned int,
                         const struct usb_device_id *);
#else
static void *ezusb_probe(struct usb_device *, unsigned int);
#endif

static void  ezusb_disconnect(struct usb_device *, void *);
static int   ezusb_download(ezusb_elem_t*);
static void  ezusb_cleanup(ezusb_elem_t*);


/* statics ********************************************************************/

static struct file_operations ezusb_procfs_fops;
static struct inode_operations ezusb_procfs_iops;

static const char *const ezusb_proc_dir = "ezusb";
                         /* directory name in /proc */

static const char *const ezusb_proc_filename = "dev";
                         /* filname of files in /proc */


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
#if EZUSB_DEVICE_TABLE
static struct usb_device_id ezusb_ids [] = 
{
        {USB_DEVICE(ANCHOR_CHIPS_VENID, AN2131_DEVID)},
        { }                                            /* Terminating entry */
};
 
MODULE_DEVICE_TABLE (usb, ezusb_ids);
#endif
#endif


static struct usb_driver ezusb_driver =
{
   name:       EZUSB_MOD_NAME,
   probe:      ezusb_probe,
   disconnect: ezusb_disconnect,
   fops:       NULL,
   minor:      0,

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
#if EZUSB_DEVICE_TABLE
   id_table:   ezusb_ids,
#endif
#endif

};

static ezusb_t  ezusb_dev;
static ezusb_t *pezusb_dev = &ezusb_dev;

static struct proc_dir_entry *pezusb_proc_root = (struct proc_dir_entry *)0;

static int number_of_connects=0;            /* total number of ezusb device */
                                            /* connects, used with          */
                                            /* ignore_renum_modula          */


/* functions ******************************************************************/

/*----------------------------------------------------------------------------*\
| if something goes wrong this informs the user; for example, if there is      |
| not a terminating intel hex record in the file being downloaded.             |
\*----------------------------------------------------------------------------*/
static void 
ezusb_load_timeout(unsigned long data)
{
   unsigned long flags = 0;
   ezusb_elem_t *pezusb;
   pezusb = (ezusb_elem_t*) data;

   if (pezusb == (ezusb_elem_t*)0)
   {
      printk(EZUSB_LL "%s: ERROR NULL pointer passed to timeout, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      return;
   }

   printk(EZUSB_LL 
          "%s: ERROR timeout on ezusb device, /proc/%s/%s%u, %s,%d\n",
          EZUSB_MOD_NAME, 
          ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num, 
          __FILE__, __LINE__);

   EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);
   if (pezusb->state & EZUSB_DOWNLOADING)
   {
      printk(EZUSB_LL "%s: short hex file maybe???, %s,%d\n",
             EZUSB_MOD_NAME,
             __FILE__, __LINE__);
   }
   pezusb->state |= EZUSB_DOWNLOAD_TIMEOUT;
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);

   return;
} 


/*----------------------------------------------------------------------------*\
| setup the file in the /proc file system for downloading hex records to the   |
| usb device; the name is based on the assigned usb device number              |
\*----------------------------------------------------------------------------*/
static int
setup_ezusb_procfs(ezusb_elem_t *pezusb)
{
   int i;
   
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
   char *pfs_name;
   /* make sure we have enough room for the name; sizeof(int)*3 because */
   /* no more than three chars per byte (0xFF == 255, 0xFFFF == 65535)  */
   pfs_name = (char*)kmalloc(strlen(ezusb_proc_filename) +
                                    (sizeof(unsigned int)*3) + 1,
                                    GFP_KERNEL);
   if (pfs_name == (char*)0)
   {
      return(-ENOMEM);
   }
 
   sprintf((char*)(pfs_name), "%s%u",
           ezusb_proc_filename, pezusb->device_num);
 
   pezusb->procfs_entry = create_proc_entry(pfs_name, EZUSB_PROCFS_PERMS,
                                            pezusb_proc_root);
   kfree(pfs_name);
 
   if (pezusb->procfs_entry == (struct proc_dir_entry*)0)
   {
      printk(EZUSB_LL "%s: could not register /proc/%s/%s%u, err=%d, %s,%d\n",
             EZUSB_MOD_NAME,
             ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num,
             i, __FILE__, __LINE__);
      return(-1);
   }

   pezusb->procfs_entry->owner = THIS_MODULE;

   pezusb->procfs_entry->uid = uid;
   pezusb->procfs_entry->gid = 0;

   pezusb->procfs_entry->proc_iops = &ezusb_procfs_iops;
   pezusb->procfs_entry->proc_fops = &ezusb_procfs_fops;

#else

   pezusb->procfs_entry.low_ino = 0;

   /* make sure we have enough room for the name; sizeof(int)*3 because */
   /* no more than three chars per byte (0xFF == 255, 0xFFFF == 65535)  */
   pezusb->procfs_entry.name = (char*)kmalloc(strlen(ezusb_proc_filename) +
                                              (sizeof(unsigned int)*3) + 1, 
                                              GFP_KERNEL);
   if (pezusb->procfs_entry.name == (char*)0)
   {
      return(-ENOMEM);
   }

   sprintf((char*)(pezusb->procfs_entry.name), "%s%u", 
           ezusb_proc_filename, pezusb->device_num);

   pezusb->procfs_entry.namelen = strlen(pezusb->procfs_entry.name);

   pezusb->procfs_entry.mode = EZUSB_PROCFS_PERMS;
   pezusb->procfs_entry.nlink = 1;
   pezusb->procfs_entry.uid = uid;
   pezusb->procfs_entry.gid = 0;
   pezusb->procfs_entry.ops = &ezusb_procfs_iops;

   /* create /proc/... file for reading code to download to board */
   if ((i=proc_register(pezusb_proc_root, &pezusb->procfs_entry)) != 0)
   {
      printk(EZUSB_LL "%s: could not register /proc/%s/%s%u, err=%d, %s,%d\n",
             EZUSB_MOD_NAME,
             ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num,
             i, __FILE__, __LINE__);
      kfree(pezusb->procfs_entry.name);
      return(i);
   }
#endif

   return(0);
}


/*----------------------------------------------------------------------------*\
| add a new ezusb element to the list and return ptr to the new element in the |
| list or NULL on error                                                        |
\*----------------------------------------------------------------------------*/
static ezusb_elem_t*
alloc_ezusb_list_element(void)
{
   unsigned long flags = 0;
   ezusb_elem_t *pezusb;

   /* allocate the element */
   pezusb = (ezusb_elem_t*) kmalloc(sizeof(ezusb_elem_t), GFP_KERNEL);
   if (pezusb == (ezusb_elem_t*)0)
   {
      return((ezusb_elem_t*)0);
   }
   memset(pezusb, 0, sizeof(ezusb_elem_t));

   init_MUTEX(&pezusb->mutex);
   init_waitqueue_head(&pezusb->remove_ok);

   init_timer(&pezusb->download_timer);
   pezusb->download_timer.data = (unsigned long) pezusb;
   pezusb->download_timer.function = ezusb_load_timeout;


   /* link the element */
   EZUSB_WRITE_LOCK(&pezusb_dev->list_lock, flags);
   pezusb->next = pezusb_dev->ezusb_list;
   pezusb_dev->ezusb_list = pezusb;
   EZUSB_WRITE_UNLOCK(&pezusb_dev->list_lock, flags);

   return(pezusb);
}


/*----------------------------------------------------------------------------*\
| delete the given element from the list                                       |
\*----------------------------------------------------------------------------*/
static void
free_ezusb_list_element(ezusb_elem_t *pezusb)
{
   unsigned long flags = 0;
   ezusb_elem_t *prev_ezusb;
   ezusb_elem_t *current_ezusb;

   if (pezusb == (ezusb_elem_t*)0)
   {
      return;
   }

   EZUSB_WRITE_LOCK(&pezusb_dev->list_lock, flags);

   prev_ezusb = pezusb_dev->ezusb_list;
   current_ezusb = pezusb_dev->ezusb_list;
   
   /* search the list for the given element */
   while ((current_ezusb != (ezusb_elem_t*)0) && 
          (current_ezusb != pezusb))
   {
      prev_ezusb = current_ezusb;
      current_ezusb = current_ezusb->next;
   }

   if (current_ezusb == (ezusb_elem_t*)0)
   {
      /* not found */
      printk(EZUSB_LL "%s: ERROR could not find struct to free, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);

      EZUSB_WRITE_UNLOCK(&pezusb_dev->list_lock, flags);

      return;
   }

   if (current_ezusb == prev_ezusb)
   {
      /* free first element in list */
      pezusb_dev->ezusb_list = current_ezusb->next;
   }
   else
   {
      /* free element in list located after the first elem */
      prev_ezusb->next = current_ezusb->next;
   }

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
   kfree(current_ezusb->procfs_entry.name);
#endif

   kfree(current_ezusb);

   EZUSB_WRITE_UNLOCK(&pezusb_dev->list_lock, flags);

   return;
}


/*----------------------------------------------------------------------------*\
| this routine is called to free the ezusb element if it is no longer in use;  |
| we must gaurd against the stuct being removed while the procfs file is open  |
| and the usb device is disconnected; the last one out (proc file close or     |
| disconnect) will free the struct.                                            |
\*----------------------------------------------------------------------------*/
static void 
ezusb_cleanup(ezusb_elem_t *pezusb)
{

   int i;
   unsigned long flags = 0;

   EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);

   if (!(pezusb->state & EZUSB_REMOVE_PENDING) ||
       (pezusb->state & EZUSB_PROC_FILE_OPENED))
   {
      /* struct may be in use so don't free it yet, let the last one   */
      /* to use the struct free it (disconnect or proc close)          */
      EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
      return;
   }

   /* if the download timer is running, stop it - it is safe because the     */
   /* procfs file is closed so we won't be getting anymore writes from it... */
   if (pezusb->state & EZUSB_DOWNLOADING)
   {
      del_timer(&pezusb->download_timer);
   }

   /* ok to free struct */
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);

   /* remove procfs file */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
   remove_proc_entry(pezusb->procfs_entry->name, pezusb_proc_root);

#else
   if (i=proc_unregister(pezusb_proc_root, pezusb->procfs_entry.low_ino))
   {
      printk(EZUSB_LL 
          "%s: could not unregister /proc/%s/%s%u, state=%#x, err=%d, %s,%d\n",
          EZUSB_MOD_NAME, 
          ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num, 
          pezusb->state, i, 
          __FILE__, __LINE__);

      /* bad news... fall through and free the memory anyway... no, */
      /* something else may be using it, just leave it, better to   */
      /* have a small memory leak than a hung system...             */

      /* dec use count so module can be unloaded.... */
      MOD_DEC_USE_COUNT;
      return;
   }
#endif

   printk(EZUSB_LL "%s: /proc/%s/%s%u disconnected\n", 
          EZUSB_MOD_NAME, 
          ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num);

   free_ezusb_list_element(pezusb);

   MOD_DEC_USE_COUNT;
   return;
}


/*----------------------------------------------------------------------------*\
| find and return the ezusb struct corresponding to the given file struct      |
| or null if not found                                                         |
\*----------------------------------------------------------------------------*/
static ezusb_elem_t*
find_ezusb_element(struct file *pfile)
{
   unsigned long flags = 0;
   ezusb_elem_t *pezusb;

   EZUSB_READ_LOCK(&pezusb_dev->list_lock, flags);

   pezusb = pezusb_dev->ezusb_list;

   /* find ezusb struct in list based on procfs file */
   while (pezusb != (ezusb_elem_t*)0)
   {
      if (pezusb->procfs_file == pfile)
      {
         break;
      }
      pezusb = pezusb->next;
   }

   EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, flags);
   return(pezusb);
}


/*----------------------------------------------------------------------------*\
| save the file struct associated with the procfs inode; make sure the file    |
| is not being re-opened; set the state to file opened so it won't be removed  |
| if the device is diconnected while the file is opened and to insure the      |
| file is not re-opened.                                                       |
\*----------------------------------------------------------------------------*/
int
ezusb_open_procfs(struct inode *pinode, struct file *pfile)
{
   unsigned long list_flags = 0;
   unsigned long state_flags = 0;
   ezusb_elem_t *pezusb;

   EZUSB_READ_LOCK(&pezusb_dev->list_lock, list_flags);
   pezusb = pezusb_dev->ezusb_list;

   /* find procfs file being opened using inode */
   while (pezusb != (ezusb_elem_t*)0)
   {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
      if (pezusb->procfs_entry->low_ino == pinode->i_ino)
#else
      if (pezusb->procfs_entry.low_ino == pinode->i_ino)
#endif
      {
         /* store file name to find struct based on file struct */
         pezusb->procfs_file = pfile;
         break;
      }
      pezusb = pezusb->next;
   }


   if (pezusb == (ezusb_elem_t*)0)
   {
      EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, list_flags);
      printk(EZUSB_LL "%s: ERROR could not find procfs inode %d, %s,%d\n",
             EZUSB_MOD_NAME, pinode->i_ino, __FILE__, __LINE__);
      return(-1);
   }

   EZUSB_SPIN_LOCK(&pezusb->state_lock, state_flags);
   if ((pezusb->state & EZUSB_REMOVE_PENDING)   ||
       (pezusb->state & EZUSB_PROC_FILE_OPENED) ||
       (pezusb->state & EZUSB_DOWNLOADING))
   {
      /* don't let file be re-opened if remove is pending, file is open,  */
      /* or download in progress; if we get a short file (one that does   */
      /* not have the last record) the file will be closed but the        */
      /* download will still be in progress...                            */
      printk(EZUSB_LL 
         "%s: ERROR could open /proc/%s/%s%u, remove/download pending, %s,%d\n",
         EZUSB_MOD_NAME, 
         ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num, 
         __FILE__, __LINE__);
      EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, list_flags);
      EZUSB_SPIN_UNLOCK(&pezusb->state_lock, state_flags);
      return(-1);
   }

   pezusb->state |= EZUSB_PROC_FILE_OPENED;
   EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, list_flags);
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, state_flags);

   return(0);
}


/*----------------------------------------------------------------------------*\
| close the file; if the device has been disconnected the cleanup routine will |
| free the struct.                                                             |
\*----------------------------------------------------------------------------*/
int
ezusb_close_procfs(struct inode *pinode, struct file *pfile)
{

   unsigned long flags = 0;
   ezusb_elem_t *pezusb;


   pezusb = find_ezusb_element(pfile);

   if (pezusb == (ezusb_elem_t*)0)
   {
      printk(EZUSB_LL 
             "%s: ERROR could not struct associated with inode %d, %s,%d\n",
             EZUSB_MOD_NAME, pinode->i_ino, __FILE__, __LINE__);
      return(-1);
   }

   EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);
   pezusb->procfs_file = (struct file*)0;
   pezusb->state &= ~EZUSB_PROC_FILE_OPENED;
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
   ezusb_cleanup(pezusb);

   return(0);
}


/*----------------------------------------------------------------------------*\
| user has written to procfs, so this function reads it; read a single         |
| complete line from the procfs file at a time; each line must end with a      |
| newline char.                                                                |
\*----------------------------------------------------------------------------*/
static int
ezusb_write_procfs(struct file *pfile,
                   const char *buf,
                   size_t length,
                   loff_t *offset)
{
   int i;
   int newline = 0;
   ezusb_elem_t *pezusb;

   pezusb = find_ezusb_element(pfile);

   if (pezusb == (ezusb_elem_t*)0)
   {
      printk(EZUSB_LL "%s: ERROR could not find ezusb struct, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      return(-1);
   }

   for (i=0; i+pezusb->buffer_bytes_used<EZUSB_BUFF_LEN-1 && i<length; i++)
   {
      get_user(pezusb->buffer[i+pezusb->buffer_bytes_used], buf+i);
      if (pezusb->buffer[i+pezusb->buffer_bytes_used] == '\n')
      {
         newline++;  /* we read a newline */
         break;
      }
   }

   pezusb->buffer_bytes_used += i;
   if (!newline)
   {
      /* we don't have a complete line, don't download it until we do */
      return(i);
   }

   pezusb->buffer[pezusb->buffer_bytes_used] = '\0';

   /* download the board... */
   ezusb_download(pezusb);
   pezusb->buffer_bytes_used = 0;

   i++;  /* acount for the null after the newline */

   return(i);
}


/*----------------------------------------------------------------------------*\
| compare the memory of the ezusb with the given buffer                        |
\*----------------------------------------------------------------------------*/
static int 
compare_ezusb_mem(struct usb_device *dev, 
                  unsigned char *pbuff,
                  unsigned short addr, 
                  unsigned char  len)
{
   unsigned i;
   unsigned char *pusbmem;
   
   if (!len)
   {
      return(0);
   }

   pusbmem = (unsigned char*)kmalloc(len, GFP_KERNEL);
   if (pusbmem == (unsigned char*)0)
   {
      printk(EZUSB_LL "%s: could alloc mem, %s,%d\n", 
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      return(-1);
   }
 
   i = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
                       ANCHOR_LOAD, VEN_REQUEST_IN,
                       addr, 0, pusbmem, len, 300);
   if (i != len) 
   {
      printk(EZUSB_LL "%s: error %d, %s,%d\n", 
             EZUSB_MOD_NAME, i, __FILE__, __LINE__);
      kfree(pusbmem);
      return(-1);
   }

   if (pbuff != (unsigned char*)0)
   {
      /* compare the memory */
      for (i=0; i < len; i++)
      {
         if (pusbmem[i] != pbuff[i])
         {
            printk(EZUSB_LL "%s: ERROR ezusb memory corrupt, %s,%d\n", 
                   EZUSB_MOD_NAME, __FILE__, __LINE__);
            kfree(pusbmem);
            return(-1);
         }
      }
   }
   else
   {
      /* just dump the memory */
      printk(EZUSB_LL "%04x:", addr);
      for (i=0; i < len; i++)
      {
         printk(" %02x", pusbmem[i]);
         if ((((i+1)%8) == 0) && ((i+1) != len))
         {
            /* newline and more text */
            printk("\n");
            printk(EZUSB_LL "     ");
         }
      }
      printk("\n");
   }

   kfree(pusbmem);

   return(0);
}


/*----------------------------------------------------------------------------*\
| debug function to dump ezusb memory...                                       |
\*----------------------------------------------------------------------------*/
static int 
display_ezusb_mem(struct usb_device *dev, 
                  unsigned short addr, 
                  unsigned char  len)
{
   /* use compare function to dump memory */
   return(compare_ezusb_mem(dev, (unsigned char *)0, addr, len));
}


/*----------------------------------------------------------------------------*\
| download a hex record the the ezusb 2131; make sure it is still connected;   |
| restart download timer too.                                                  |
\*----------------------------------------------------------------------------*/
static int
ezusb_download(ezusb_elem_t *pezusb)
{
   int   i;
   unsigned long flags = 0;
   char  cmd;
   char  tmp[5];
   char *pch;
   intel_hex_record_t ihex_record;


   EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);
   if (pezusb->state & EZUSB_REMOVE_PENDING)
   {
      /* as long as we are reading lines from the procfs, keep */
      /* extending the timeout if it has been started...       */
      if (pezusb->state & EZUSB_DOWNLOADING)
      {
         mod_timer(&pezusb->download_timer, jiffies + EZUSB_LOAD_TIMEOUT);
      }
      EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
      return(-1);
   }

   if (!(pezusb->state & EZUSB_DOWNLOADING))
   {
      /* put 8051 into RESET */
      pezusb->state |= EZUSB_DOWNLOADING;
      cmd=CPUCS_RESET;
      i=usb_control_msg(pezusb->usbdev, 
                        usb_sndctrlpipe(pezusb->usbdev, 0), 
                        ANCHOR_LOAD, VEN_REQUEST_OUT, 
                        CPUCS_REG, 0, &cmd, 1, 300);
      if (i != 1)
      {
         printk(EZUSB_LL "%s: could not put 8051 into RESET, %s,%d\n", 
             EZUSB_MOD_NAME, __FILE__, __LINE__);
         EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
         return(-1);
      }
      printk(EZUSB_LL "%s: download started for /proc/%s/%s%u\n",
             EZUSB_MOD_NAME, 
             ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num);

      /* start download timer */
      pezusb->download_timer.expires = jiffies + EZUSB_LOAD_TIMEOUT;
      add_timer(&pezusb->download_timer);
   }
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);

   /* parse the line into an intel hex record */
   /* the records look like this:             */
   /* :090160001204283119740280E236           */

   if (pezusb->buffer[0] !=  ':')
   {
      printk(EZUSB_LL 
             "%s: could not find start of record \":\", line=%s, %s,%d\n", 
             EZUSB_MOD_NAME, 
             pezusb->buffer, 
             __FILE__, __LINE__);
      return(-1);
   }

   strncpy(tmp, &pezusb->buffer[1], 2);
   tmp[2] = '\0';
   ihex_record.length  = (unsigned char)simple_strtoul(tmp, (char**)0, 16);

   if (ihex_record.length > MAX_IHEX_RECORD_LEN)
   {
      printk(EZUSB_LL 
             "%s: bad length given, %#x greater than %#x, line=%s, %s,%d\n", 
             EZUSB_MOD_NAME, ihex_record.length, 
             MAX_IHEX_RECORD_LEN, pezusb->buffer,
             __FILE__, __LINE__);
      return(-1);
   }

   strncpy(tmp, &pezusb->buffer[3], 4);
   tmp[4] = '\0';
   ihex_record.address  = (unsigned short)simple_strtoul(tmp, (char**)0, 16);

   strncpy(tmp, &pezusb->buffer[7], 2);
   tmp[2] = '\0';
   ihex_record.type  = (unsigned char)simple_strtoul(tmp, (char**)0, 16);


   /* if last record then kick the board and let it re-enumerate */
   /* last record has a type of 1                                */
   if (ihex_record.type == 1)
   {
      EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);
      pezusb->state &= ~EZUSB_DOWNLOADING;

      /* stop download timer */
      del_timer(&pezusb->download_timer);

      /* take the 8051 out of RESET */
      cmd=CPUCS_RUN;
      i=usb_control_msg(pezusb->usbdev, 
                        usb_sndctrlpipe(pezusb->usbdev, 0), 
                        ANCHOR_LOAD, VEN_REQUEST_OUT, 
                        CPUCS_REG, 0, &cmd, 1, 300);
      if (i != 1)
      {
         printk(EZUSB_LL "%s: could not take 8051 out of RESET, %s,%d\n", 
                EZUSB_MOD_NAME, __FILE__, __LINE__);
         EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
         return(-1);
      }

      printk(EZUSB_LL 
             "%s: downloaded %d bytes to /proc/%s/%s%u, re-enumerating\n",
             EZUSB_MOD_NAME, 
             pezusb->bytes_downloaded,
             ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num);


      EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
      return(0);
   }
   else if (ihex_record.type != 0)
   {
     /* just print warning to inform user - don't stop the download.... */
      printk(EZUSB_LL 
         "%s: WARNING /proc/%s/%s%u maybe problem record type \"%s\", %s,%d\n", 
         EZUSB_MOD_NAME, 
         ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num,
         pezusb->buffer, __FILE__, __LINE__);
   }

   tmp[2] = '\0';
   pch = &pezusb->buffer[9];  /* start of hex data */
   for (i=0; i<ihex_record.length; i++, pch+=2)
   {
      strncpy(tmp, pch, 2);
      ihex_record.data[i]  = (unsigned char)simple_strtoul(tmp, (char**)0, 16);
   }

   /* sanity check, we want to help out as much as we can; if we went past */
   /* the record length, print a warning but don't stop the download...    */
   /* note: need to watch out for carriage return before newline in some   */
   /*       files.                                                         */
   i = (isxdigit(pezusb->buffer[pezusb->buffer_bytes_used-1]))?
                 pezusb->buffer_bytes_used-2:   /* no carriage return */
                 pezusb->buffer_bytes_used-3;   /* carriage return    */

   if (pch != &pezusb->buffer[i])  /* both ptrs should be at checksum */
   {
      printk(EZUSB_LL 
           "%s: WARNING /proc/%s/%s%u maybe problem with line \"%s\", %s,%d\n", 
           EZUSB_MOD_NAME, 
           ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num,
           pezusb->buffer, __FILE__, __LINE__);
   }

   /* download the line of code to the board */
   i=usb_control_msg(pezusb->usbdev, 
                     usb_sndctrlpipe(pezusb->usbdev, 0), 
                     ANCHOR_LOAD, VEN_REQUEST_OUT, 
                     ihex_record.address, 0, ihex_record.data, 
                     ihex_record.length, 300);
   if (i != ihex_record.length)
   {
      printk(EZUSB_LL "%s: ERROR could not download board, err=%d, %s,%d\n", 
             EZUSB_MOD_NAME, i, __FILE__, __LINE__);
      return(-1);
   }
   pezusb->bytes_downloaded += ihex_record.length;

   /* make sure ezusb memory was downloaded correctly */
   if (compare_ezusb_mem(pezusb->usbdev,
                         ihex_record.data,
                         ihex_record.address,
                         ihex_record.length))
   {
      unsigned u;
      /* dump the line from the board for debugging */
      display_ezusb_mem(pezusb->usbdev, 
                        ihex_record.address, 
                        ihex_record.length);

      printk(EZUSB_LL "line:");
      for (u = 0; u < ihex_record.length; u++)
      {
         printk(EZUSB_LL " %02x", ihex_record.data[u]);
         if ((((u+1)%8) == 0) && ((u+1) != ihex_record.length))
         {
            /* newline and more text */
            printk("\n");
            printk(EZUSB_LL "     ");
         }
      }
      printk("\n");
   }

   return(0);
}


/*----------------------------------------------------------------------------*\
| a basic probe for ezusb devices; called for any unclaimed usb devices at     |
| load time and when any new usb device is connected; alloc struct and         |
| create file in /proc for downloading...                                      |
\*----------------------------------------------------------------------------*/
static void*
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
ezusb_probe(struct usb_device *usbdev, unsigned int interface_num,
            const struct usb_device_id *id)
#else
ezusb_probe(struct usb_device *usbdev, unsigned int interface_num)
#endif
{
   int i;
   unsigned long flags = 0;
   ezusb_elem_t *pezusb;

   /* see if we have an Anchor Chips EZ-USB AN2131Q device */
   if (usbdev->descriptor.idVendor  != vendor_id || 
       usbdev->descriptor.idProduct != product_id)
   {
      return(NULL);
   }

   printk(EZUSB_LL "%s: found ezusb venid %#x, prodid %#x interface_num:%d\n",
          EZUSB_MOD_NAME, usbdev->descriptor.idVendor, 
          usbdev->descriptor.idProduct, interface_num);

   if (interface_num != 0)
   {
      /* we only support interface 0 to download the board, print message */
      /* for debugging purposes                                           */
      printk(EZUSB_LL "%s: interface number not zero, %d, %s,%d\n",
             EZUSB_MOD_NAME, interface_num, __FILE__, __LINE__);
      return(NULL);
   }

   /* we don't handle multiple configurations */
   if (usbdev->descriptor.bNumConfigurations != 1)
   {
      /* print message for debugging purposes  */
      printk(EZUSB_LL "%s: number of configs != 1, %d, %s,%d\n",
             EZUSB_MOD_NAME, usbdev->descriptor.bNumConfigurations, 
             __FILE__, __LINE__);
      return(NULL);
   }

   if (!multi_boards)
   {
      /* user only needs single board support          */
      EZUSB_READ_LOCK(&pezusb_dev->list_lock, flags);
      if (pezusb_dev->ezusb_list != (ezusb_elem_t*)0)
      {
         EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, flags);
         printk(EZUSB_LL "%s: ERROR set multi_boards module parameter\n",
                EZUSB_MOD_NAME);
         printk(EZUSB_LL "for multiple board support, %s,%d\n",
                __FILE__, __LINE__);
         return(NULL);
      }
      EZUSB_READ_UNLOCK(&pezusb_dev->list_lock, flags);
   }

   /* the ignore_renum_modula hack inspired by      */
   /* Claes Nygren (claes * wijnet_se)              */
   /* ok, so I overload the list_lock a bit here... */
   EZUSB_WRITE_LOCK(&pezusb_dev->list_lock, flags);
   number_of_connects++;            /* total number of ezusb device */
                                    /* connects                     */
   if (number_of_connects%ignore_renum_modula == 0)
   {
      /* when number_of_connects goes negative, the modula will still be */
      /* correct (but negative) so don't worry about it going negative.  */
      EZUSB_WRITE_UNLOCK(&pezusb_dev->list_lock, flags);
      printk(EZUSB_LL "%s: IGNORING device due to ignore_renum_modula, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      return(NULL);
   }
   EZUSB_WRITE_UNLOCK(&pezusb_dev->list_lock, flags);

   /* find a free location */
   pezusb = alloc_ezusb_list_element();
   if (pezusb == (ezusb_elem_t*)0)
   {
      printk(EZUSB_LL "%s: ERROR out of memory, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      return(NULL);
   }

   pezusb->usbdev = usbdev;
   pezusb->state = 0;
   if (multi_boards)
   {
      pezusb->device_num = usbdev->devnum;
   }
   else
   {
      pezusb->device_num = 0;
   }

   if (setup_ezusb_procfs(pezusb))
   {
      printk(EZUSB_LL "%s: could not setup procfs for download, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      free_ezusb_list_element(pezusb);
      return(NULL);
   }

   if (usb_set_configuration(usbdev, usbdev->config[0].bConfigurationValue) < 0)
   {
      printk(EZUSB_LL "%s: set_configuration failed, %s,%d\n",
             EZUSB_MOD_NAME, __FILE__, __LINE__);
      free_ezusb_list_element(pezusb);
      return(NULL);
   }

   spin_lock_init(&pezusb->state_lock);

   printk(EZUSB_LL "%s: ready for download from /proc/%s/%s%u\n", 
          EZUSB_MOD_NAME, 
          ezusb_proc_dir, ezusb_proc_filename, pezusb->device_num);

   MOD_INC_USE_COUNT;

   return(pezusb);
}


/*----------------------------------------------------------------------------*\
| one of our registered ezusb devices has been disconnected or maybe it has    |
| been downloaded and is re-enumerating; call the cleanup routine to free      |
| structures.                                                                  |
\*----------------------------------------------------------------------------*/
static void 
ezusb_disconnect(struct usb_device *usbdev, void *drv_context)
{
   int i;
   unsigned long flags = 0;
   ezusb_elem_t *pezusb;

   pezusb = (ezusb_elem_t*) drv_context;

   EZUSB_SPIN_LOCK(&pezusb->state_lock, flags);
   pezusb->state |= EZUSB_REMOVE_PENDING;
   EZUSB_SPIN_UNLOCK(&pezusb->state_lock, flags);
   ezusb_cleanup(pezusb);

   return;
}


/*----------------------------------------------------------------------------*\
| the basic init; setup the directory in /proc to allow connected boards to    |
| be downloaded.                                                               |
\*----------------------------------------------------------------------------*/
int __init 
ezusb_module_init(void)
{
   int i;

   /* log copyright per GPL guidelines; warning level is used */
   /* to make sure it gets out                               */
   printk(KERN_WARNING EZUSB_DESCRIPTION ", Version " 
                       EZUSB_VERSION"; " EZUSB_COPYRIGHT"\n");

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
   ezusb_procfs_iops.default_file_ops = &ezusb_procfs_fops;
#endif

   ezusb_procfs_fops.write = ezusb_write_procfs;
   ezusb_procfs_fops.open = ezusb_open_procfs;
   ezusb_procfs_fops.release = ezusb_close_procfs;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
   pezusb_proc_root = (struct proc_dir_entry*)
                      proc_mkdir(ezusb_proc_dir,
                                 (struct proc_dir_entry *)0);
#else
   pezusb_proc_root = (struct proc_dir_entry*)
                      create_proc_entry(ezusb_proc_dir, S_IFDIR, 
                                        (struct proc_dir_entry *)0);
#endif

   if (pezusb_proc_root == (struct proc_dir_entry *)0) 
   {
      printk(EZUSB_LL "%s: could not create /proc/%s directory, %s,%d\n",
             EZUSB_MOD_NAME, ezusb_proc_dir, __FILE__, __LINE__);
      return(-ENOENT);
   }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0))
   pezusb_proc_root->owner = THIS_MODULE;
#endif

   pezusb_dev->ezusb_list = (ezusb_elem_t*)0;
   pezusb_dev->list_lock = RW_LOCK_UNLOCKED;

   if (uid < 0)
   {
      printk(EZUSB_LL 
             "%s: WARNING bad uid parameter (%d), using uid of 0, %s,%d\n", 
             EZUSB_MOD_NAME, uid,
             __FILE__, __LINE__);
      uid = 0;
   }

   if ((multi_boards != 0) &&
       (multi_boards != 1))
   {
      /* if user sets multi_boards, it is probably for multiple board support */
      printk(EZUSB_LL 
             "%s: WARNING bad multi_boards parameter (%d), using 1, %s,%d\n", 
             EZUSB_MOD_NAME, multi_boards,
             __FILE__, __LINE__);
      multi_boards = 1;
   }

   if (ignore_renum_modula <= 1)
   {
      /* don't allow bad parms... a value of 1 would cause all connects to */
      /* be ignored...                                                     */
      printk(EZUSB_LL 
        "%s: WARNING bad ignore_renum_modula parameter (%d), using %d, %s,%d\n",
             EZUSB_MOD_NAME, ignore_renum_modula, INT_MAX,
             __FILE__, __LINE__);
      ignore_renum_modula = INT_MAX;
   }

   if (ignore_renum_modula != INT_MAX)
   {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
#if EZUSB_DEVICE_TABLE
      /* ignore_renum_modula will not work if EZUSB_DEVICE_TABLE is true.     */
      /* TODO: find out why this is so... it seems we get called twice        */
      /*       when EZUSB_DEVICE_TABLE is set and the ignore_renum_modula     */
      /*       code path ignores the connect... this is shown below in the    */
      /*       messages file; see, we are probed twice to claim the device.   */
      /*       until this is resolved, don't allow EZUSB_DEVICE_TABLE to be   */
      /*       true when ignore_renum_modula is set. the modula is set to 2   */
      /*       for the text below, and the second and third connects are      */
      /*       shown below.                                                   */
      /*                                                                      */
      /* usb.c: USB disconnect on device 32                                   */
      /* ezusb2131: /proc/ezusb/dev0 disconnected                             */
      /* hub.c: USB new device connect on bus1/1/1, assigned device number 33 */
      /* ezusb2131: found ezusb venid 0x547, prodid 0x2131 interface_num:0    */
      /* ezusb2131: IGNORING device due to ignore_renum_modula,               */
      /*            ezusb2131.c,1202                                          */
      /* ezusb2131: found ezusb venid 0x547, prodid 0x2131 interface_num:0    */
      /* ezusb2131: ready for download from /proc/ezusb/dev0                  */
      /*  /etc/hotplug/usb.agent: ... no drivers for USB product 547/2131/4   */

      printk(KERN_ERR 
             "%s: %s; %s %s to 0, %s,%d\n",
             EZUSB_MOD_NAME, 
             "ERROR: can't have EZUSB_DEVICE_TABLE and ignore_renum_modula set",
             "you must set the #define EZUSB_DEVICE_TABLE in",
             __FILE__,
             __FILE__, __LINE__);
      remove_proc_entry(ezusb_proc_dir, 0);
      return(-1);

#endif
#endif

      printk(EZUSB_LL 
           "%s: ignore_renum_modula is being used and is set to %d,  %s,%d\n",
             EZUSB_MOD_NAME, ignore_renum_modula,
             __FILE__, __LINE__);
   }

   printk(EZUSB_LL "%s: uid %d may download %s with venid=%#x/prodid=%#x\n", 
          EZUSB_MOD_NAME, uid,
          (multi_boards)?"multiple connected boards":
                         "a single connected board",
          vendor_id, product_id);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
#if EZUSB_DEVICE_TABLE
   /* copy the module parameters into the device table */
   ezusb_ids[0].idVendor  = vendor_id;
   ezusb_ids[0].idProduct = product_id;
#endif
#endif

   if (i=usb_register(&ezusb_driver))
   {
      printk(EZUSB_LL "%s: usb_register call failed, err=%d, %s,%d\n",
             EZUSB_MOD_NAME, i, __FILE__, __LINE__);
      remove_proc_entry(ezusb_proc_dir, 0);
      return(-1);
   }

   return(0);
}


/*----------------------------------------------------------------------------*\
| the basic cleanup; remove the ezusb download directory in /proc; the mod     |
| count prevents this cleanup routine from being called if there are devices   |
| connected to it (and thus files in the directory)                            |
\*----------------------------------------------------------------------------*/
void __exit 
ezusb_module_exit(void)
{
   remove_proc_entry(ezusb_proc_dir, 0);
   usb_deregister(&ezusb_driver);
}

module_init(ezusb_module_init);
module_exit(ezusb_module_exit);

